# future
from __future__ import annotations

# stdlib
import socket
from typing import Any
from typing import Callable
from typing import Dict as TypeDict
from typing import List as TypeList
from typing import Optional
from typing import Tuple as TypeTuple
from typing import Union

# relative
from .deps import allowed_hosts
from .lib import find_available_port

ALLOWED_NODE_TYPES = ["domain", "network", "gateway", "enclave"]


class BadGrammar(Exception):
    pass


class GrammarVerb:
    def __init__(
        self,
        command: str,
        full_sentence: TypeList[TypeDict[str, Any]],
        abbreviations: TypeDict[int, TypeList[Optional[str]]],
    ) -> None:
        self.grammar: TypeList[
            Union[GrammarTerm, HostGrammarTerm, SourceGrammarTerm]
        ] = []
        self.command = command
        self.full_sentence = full_sentence
        self.abbreviations = abbreviations

    def get_named_term_grammar(self, name: str) -> GrammarTerm:
        for term in self.grammar:
            if term.name == name and isinstance(term, GrammarTerm):
                return term
        raise BadGrammar(f"GrammarTerm with {name} not found in {self.grammar}")

    def get_named_term_hostgrammar(self, name: str) -> HostGrammarTerm:
        for term in self.grammar:
            if term.name == name and isinstance(term, HostGrammarTerm):
                return term
        raise BadGrammar(f"HostGrammarTerm with {name} not found in {self.grammar}")

    def get_named_term_type(
        self, name: str, term_type: Optional[str] = None
    ) -> Union[GrammarTerm, HostGrammarTerm]:
        if term_type == "host":
            return self.get_named_term_hostgrammar(name=name)
        return self.get_named_term_grammar(name=name)

    def set_named_term_type(
        self, name: str, new_term: GrammarTerm, term_type: Optional[str] = None
    ) -> None:
        new_grammar = []
        for term in self.grammar:
            found = False
            if term.name == name:
                if term_type is not None and term.type == term_type:
                    found = True
                elif term_type is None:
                    found = True
            if not found:
                new_grammar.append(term)
            else:
                new_grammar.append(new_term)
        self.grammar = new_grammar

    def load_grammar(
        self, grammar: TypeList[Union[GrammarTerm, HostGrammarTerm, SourceGrammarTerm]]
    ) -> None:
        self.grammar = grammar


class GrammarTerm:
    def __init__(
        self,
        type: str,
        name: str,
        default: Optional[Union[str, Callable]] = None,
        options: Optional[TypeList] = None,
        example: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        self.raw_input: Optional[str] = None
        self.input: Optional[str] = None
        self.type = type
        self.name = name
        self.default = default
        self.options = options if options is not None else []
        self.example = example

    @property
    def snake_input(self) -> Optional[str]:
        if self.input:
            return self.input.lower().replace(" ", "_")
        return None

    @property
    def kebab_input(self) -> Optional[str]:
        if self.input:
            return self.input.lower().replace(" ", "-")
        return None

    def __repr__(self) -> str:
        return f"<{type(self).__name__}: {self.name}<{self.type}>: {self.input} [raw: {self.raw_input}]>"

    def get_example(self) -> str:
        return_value = self.example if self.example else self.default
        if callable(return_value):
            return_value = return_value()
        return str(return_value)

    # no op
    def custom_parsing(self, input: str) -> str:
        return input

    def parse_input(self, input: Optional[str]) -> None:
        self.raw_input = input
        if input is None and self.default is None:
            raise BadGrammar(
                f"{self.name} has no default, please use one of the following options: {self.options}"
            )
        if input is None:
            if isinstance(self.default, str):
                input = self.default
            elif callable(self.default):
                input = self.default()

        if len(self.options) > 0 and input not in self.options:
            raise BadGrammar(
                f"{input} is not valid for {self.name} please use one of the following options: {self.options}"
            )

        self.input = self.custom_parsing(input=input) if input else input


class HostGrammarTerm(GrammarTerm):
    @property
    def host(self) -> Optional[str]:
        return self.parts()[0]

    @property
    def port(self) -> Optional[int]:
        return self.parts()[1]

    @property
    def search(self) -> bool:
        return bool(self.parts()[2])

    @property
    def port_tls(self) -> int:
        if self.port == 80:
            return 443
        return 444

    @property
    def free_port(self) -> int:
        if self.port is None:
            raise BadGrammar(
                f"{type(self)} unable to check if port {self.port} is free"
            )
        return find_available_port(host="localhost", port=self.port, search=self.search)

    @property
    def free_port_tls(self) -> int:
        if self.port_tls is None:
            raise BadGrammar(
                f"{type(self)} unable to check if tls port {self.port_tls} is free"
            )
        return find_available_port(host="localhost", port=self.port_tls, search=True)

    def parts(self) -> TypeTuple[Optional[str], Optional[int], bool]:
        host = None
        port: Optional[int] = None
        search = False
        if self.input:
            parts = self.input.split(":")
            host = parts[0]
            if len(parts) > 1:
                port_str = parts[1]
                if port_str.endswith("+"):
                    search = True
                    port_str = port_str[0:-1]
                port = int(port_str)
        return (host, port, search)

    def validate_host(self, host_or_ip: str) -> bool:
        try:
            if socket.gethostbyname(host_or_ip) == host_or_ip:
                return True
            elif socket.gethostbyname(host_or_ip) != host_or_ip:
                return True
        except socket.gaierror:
            raise BadGrammar(
                f"{host_or_ip} is not valid for {self.name}. Try an IP, hostname or docker, vm, aws, azure or gcp"
            )
        return False

    def validate_port(self, port: str) -> bool:
        try:
            if port.endswith("+"):
                int(port[0:-1])
            else:
                int(port)
        except Exception:  # nosec
            raise BadGrammar(
                f"{port} is not a valid port option. Try: {self.get_example()}"
            )
        return True

    def custom_parsing(self, input: str) -> str:
        colons = input.count(":")
        host = input
        port = None
        if colons > 1:
            raise BadGrammar(
                f"You cannot have more than one : for {self.name}, try: {self.get_example()}"
            )
        elif colons == 1:
            parts = input.split(":")
            host = parts[0]
            port = parts[1]

        if port is None:
            if host == "docker":
                port = "8081+"  # default
            else:
                port = "80"  # default

        if host not in allowed_hosts:
            _ = self.validate_host(host_or_ip=host)

        _ = self.validate_port(port=port)

        return f"{host}:{port}"


class SourceGrammarTerm(GrammarTerm):
    def custom_parsing(self, input: str) -> str:
        trimmed = input
        if trimmed.startswith("http://"):
            trimmed = trimmed.replace("http://", "")
        if trimmed.startswith("https://"):
            trimmed = trimmed.replace("https://", "")
        if trimmed.startswith("github.com/"):
            trimmed = trimmed.replace("github.com/", "")

        parts = trimmed.split("/")
        if "tree" not in input or len(parts) < 4:
            raise BadGrammar(
                f"{self.name} should be a valid github.com repo branch url. Try: {self.get_example()}"
            )

        repo = f"{parts[0]}/{parts[1]}"
        branch = "/".join(parts[3:])

        return f"{repo}:{branch}"


def validate_arg_count(arg_count: int, verb: GrammarVerb) -> bool:
    valid = True

    if arg_count not in verb.abbreviations:
        error_str = f"Command {verb.command} supports the following invocations:\n"
        for count in sorted(verb.abbreviations.keys()):
            abbreviation = verb.abbreviations[count]
            example_terms = []
            for i, term_type in enumerate(abbreviation):
                if term_type is not None:
                    term_settings = verb.full_sentence[i]
                    example = term_settings["klass"](**term_settings).get_example()
                    example_terms.append(example)
            error_str += f"{count} args: {verb.command} {' '.join(example_terms)}\n"

        raise BadGrammar(error_str)

    return valid


def launch_shorthand_support(args: TypeTuple) -> TypeTuple:
    """When launching, we want to be able to default to 'domain' if it's not provided, to launch
    nodes when no name is provided, and to support node names which have multiple words.

    hagrid launch -> hagrid launch domain
    hagrid launch United Nations -> hagrid launch "United Nations" domain
    hagrid launch United Nations domain -> hagrid launch "United Nations" domain
    hagrid launch on docker -> hagrid launch domain on docker

    """

    # Some mild analysis
    found_node_type = False
    preposition_position = 10000
    for i, arg in enumerate(args):
        if arg in ALLOWED_NODE_TYPES:
            found_node_type = True

        if arg.strip() in ["to", "from"]:
            if i < preposition_position:
                preposition_position = i

    _args = list(args)

    # Default to domain if it's not provided
    if not found_node_type:
        if preposition_position != 10000:
            _args.insert(preposition_position, "domain")
            preposition_position += 1
        else:
            _args = _args + ["domain"]

    # if there are no prepositions and the domain/network is the last word
    if preposition_position == 10000 and _args[-1] in ALLOWED_NODE_TYPES:
        _args = [" ".join(_args[:-1])] + _args[-1:]

    # if there are prepositions then combine the words in the name if there are multiple
    elif preposition_position != 10000:
        _args = [" ".join(_args[: preposition_position - 1])] + _args[
            preposition_position - 1 :
        ]

    # if there wasn't a name provided - make sure we don't have an empty place in the list
    # so that later logic will generate a name
    if _args[0] == "":
        _args = _args[1:]

    args = tuple(_args)

    return args


def parse_grammar(args: TypeTuple, verb: GrammarVerb) -> TypeList[GrammarTerm]:
    # if the command is a launch, check if any shorthands were employed
    if verb.command == "launch":
        args = launch_shorthand_support(args=args)

    arg_list = list(args)
    arg_count = len(arg_list)
    errors = []
    if validate_arg_count(arg_count=arg_count, verb=verb):
        terms = []
        abbreviation = verb.abbreviations[arg_count]
        for i, term_type in enumerate(abbreviation):
            if term_type is None:
                arg = None  # use None so we get the default
            else:
                arg = arg_list.pop(0)  # use a real arg

            term_settings = verb.full_sentence[i]

            try:
                term = term_settings["klass"](**term_settings)
                term.parse_input(arg)
                terms.append(term)

            except BadGrammar as e:
                errors.append(str(e))

        if len(errors) > 0:
            raise BadGrammar("\n".join(errors))

        # make command
        return terms
    else:
        raise BadGrammar("Grammar is not valid")



# stdlib
from typing import Optional


class AuthCredentials:
    def __init__(
        self,
        username: str,
        key_path: Optional[str] = None,
        password: Optional[str] = None,
    ) -> None:
        self.username = username
        self.key_path = key_path
        self.password = password

    @property
    def uses_key(self) -> bool:
        return bool(self.username and self.key_path)

    @property
    def uses_password(self) -> bool:
        return bool(self.username and self.password)

    @property
    def valid(self) -> bool:
        return bool(self.uses_key or self.uses_password)



# stdlib
from typing import Dict as TypeDict
from typing import List as TypeList
from typing import Optional

# relative
from .grammar import GrammarTerm
from .grammar import GrammarVerb
from .grammar import HostGrammarTerm


def get_land_verb() -> GrammarVerb:
    full_sentence = [
        {
            "name": "node_name",
            "type": "adjective",
            "klass": GrammarTerm,
            "example": "'my_domain'",
        },
        {
            "name": "preposition",
            "type": "preposition",
            "klass": GrammarTerm,
            "default": "at",
            "options": ["at", "on"],
        },
        {
            "name": "host",
            "type": "propernoun",
            "klass": HostGrammarTerm,
            "default": "docker",
            "example": "docker",
        },
    ]

    abbreviations: TypeDict[int, TypeList[Optional[str]]] = {
        3: [
            "adjective",
            "preposition",
            "propernoun",
        ],  # node_name  # at  # host
        2: [
            "adjective",
            None,
            "propernoun",
        ],  # node_name  # ignore  # host
        1: [
            "adjective",
            None,
            None,
        ],  # node_name  # ignore  # ignore
        0: [
            None,
            None,
            None,
        ],  # ignore  # ignore  # ignore
    }

    return GrammarVerb(
        command="land",
        full_sentence=full_sentence,
        abbreviations=abbreviations,
    )



# stdlib
from dataclasses import dataclass
import os
from pathlib import Path
import sys
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from urllib.parse import urlparse
import zipfile

# third party
import click
import requests
from tqdm import tqdm

# relative
from .cache import DEFAULT_BRANCH
from .cache import DEFAULT_REPO
from .cache import arg_cache
from .nb_output import NBOutput

directory = os.path.expanduser("~/.hagrid/quickstart/")


def quickstart_download_notebook(
    url: str, directory: str, reset: bool = False, overwrite_all: bool = False
) -> Tuple[str, bool, bool]:
    os.makedirs(directory, exist_ok=True)
    file_name = os.path.basename(url).replace("%20", "_").replace(" ", "_")
    file_path = directory + os.sep + file_name
    file_path = os.path.abspath(file_path)

    file_exists = os.path.isfile(file_path)
    if overwrite_all:
        reset = True

    if file_exists and not reset:
        response = click.prompt(
            f"\nOverwrite {file_name}?",
            prompt_suffix="(a/y/N)",
            default="n",
            show_default=False,
        )
        if response.lower() == "a":
            reset = True
            overwrite_all = True
        elif response.lower() == "y":
            reset = True
        else:
            print(f"Skipping {file_name}")
            reset = False

    downloaded = False
    if not file_exists or file_exists and reset:
        print(f"Downloading notebook: {file_name}")
        r = requests.get(url, allow_redirects=True)  # nosec
        with open(os.path.expanduser(file_path), "wb") as f:
            f.write(r.content)
        downloaded = True
    return file_path, downloaded, overwrite_all


def fetch_notebooks_for_url(
    url: str,
    directory: str,
    reset: bool = False,
    repo: str = DEFAULT_REPO,
    branch: str = DEFAULT_BRANCH,
    commit: Optional[str] = None,
) -> List[str]:
    downloaded_files = []
    allowed_schemes_as_url = ["http", "https"]
    url_scheme = urlparse(url).scheme
    # relative mode
    if url_scheme not in allowed_schemes_as_url:
        notebooks = get_urls_from_dir(repo=repo, branch=branch, commit=commit, url=url)
        if url.endswith(".ipynb"):
            file_name = os.path.basename(url)
            url_parts = url.split("notebooks")
            if len(url_parts) > 1:
                url_dir = url_parts[-1]
            else:
                url_dir = url
            url_dir = url_dir.replace(file_name, "")
        else:
            url_dir = url
        notebook_files = []
        existing_count = 0
        for notebook_url in notebooks:
            url_filename = os.path.basename(notebook_url)
            url_dirname = os.path.dirname(notebook_url)
            if (
                url_dirname.endswith(url_dir)
                and os.path.isdir(directory + url_dir)
                and os.path.isfile(directory + url_dir + os.sep + url_filename)
            ):
                notebook_files.append(url_dir + os.sep + url_filename)
                existing_count += 1

        if existing_count > 0:
            plural = "s" if existing_count > 1 else ""
            print(
                f"You have {existing_count} existing notebook{plural} matching: {url}"
            )
            for nb in notebook_files:
                print(nb)

        overwrite_all = False
        for notebook_url in tqdm(notebooks):
            file_path, _, overwrite_all = quickstart_download_notebook(
                url=notebook_url,
                directory=os.path.abspath(directory + os.sep + str(url_dir) + os.sep),
                reset=reset,
                overwrite_all=overwrite_all,
            )
            downloaded_files.append(file_path)

    else:
        file_path, _, _ = quickstart_download_notebook(
            url=url, directory=directory, reset=reset
        )
        downloaded_files.append(file_path)
    return downloaded_files


def quickstart_extract_notebook(
    zip_file: str,
    name: str,
    directory: Path,
    reset: bool = False,
    overwrite_all: bool = False,
) -> Tuple[str, bool, bool]:
    directory.mkdir(exist_ok=True)
    reset = overwrite_all

    base_name = os.path.basename(name)
    file_path = directory / base_name
    file_name = file_path.name
    file_exists = file_path.exists()

    if file_exists and not reset:
        response = click.prompt(
            f"\nOverwrite {file_name}?",
            prompt_suffix="(a/y/N)",
            default="n",
            show_default=False,
        )
        if response.lower() == "a":
            reset = True
            overwrite_all = True
        elif response.lower() == "y":
            reset = True
        else:
            print(f"Skipping {file_name}")
            reset = False

    extracted = False
    if not file_exists or file_exists and reset:
        print(f"Extracting notebook: {file_name}")
        with zipfile.ZipFile(zip_file, "r") as zf:
            zip_info = zf.getinfo(name)
            zip_info.filename = base_name
            zf.extract(zip_info, directory)
        extracted = True
    return str(file_path.absolute()), extracted, overwrite_all


def fetch_notebooks_from_zipfile(
    path: str, directory: str, reset: bool = False
) -> List[str]:
    dir_path = Path(directory)

    with zipfile.ZipFile(path, "r") as zf:
        notebooks = [f for f in zf.namelist() if f.endswith(".ipynb")]

    notebook_files = [dir_path / os.path.basename(nb) for nb in notebooks]
    existing_files = [nb for nb in notebook_files if nb.exists()]

    existing_count = len(existing_files)

    if existing_count > 0:
        plural = "s" if existing_count > 1 else ""
        print(f"You have {existing_count} existing notebook{plural}")
        for nb in existing_files:
            print(nb)

    extracted_files = []
    overwrite_all = False
    for notebook in tqdm(notebooks):
        file_path, _, overwrite_all = quickstart_extract_notebook(
            zip_file=path,
            name=notebook,
            directory=dir_path,
            reset=reset,
            overwrite_all=overwrite_all,
        )
        extracted_files.append(file_path)

    return extracted_files


@dataclass
class Tutorial:
    filename: str
    description: str
    url: str


REPO_RAW_PATH = "https://raw.githubusercontent.com/OpenMined/PySyft"

TUTORIALS = {
    "api/0.8": Tutorial(
        filename="api/0.8",
        description="0.8 API Notebooks",
        url="api/0.8",
    ),
    "hello-syft": Tutorial(
        filename="tutorials/hello-syft",
        description="Hello Syft",
        url="tutorials/hello-syft",
    ),
    "data-engineer": Tutorial(
        filename="tutorials/data-engineer",
        description="Data Engineer",
        url="tutorials/data-engineer",
    ),
    "data-owner": Tutorial(
        filename="tutorials/data-owner",
        description="Data Owner",
        url="tutorials/data-owner",
    ),
    "data-scientist": Tutorial(
        filename="tutorials/data-scientist",
        description="Data Scientist",
        url="tutorials/data-scientist",
    ),
    "pandas-cookbook": Tutorial(
        filename="tutorials/pandas-cookbook",
        description="Pandas Cookbook",
        url="tutorials/pandas-cookbook",
    ),
}


class QuickstartUI:
    @property
    def tutorials(self) -> Dict[str, Tutorial]:
        return TUTORIALS

    def download(
        self, tutorial_name: str, reset: bool = False, branch: str = "dev"
    ) -> NBOutput:
        if tutorial_name not in TUTORIALS.keys():
            return NBOutput(
                f'<div class="alert alert-danger">{tutorial_name} is not a valid tutorial name.</div>'
            )
        else:
            tutorial = TUTORIALS[tutorial_name]
            downloaded_files = fetch_notebooks_for_url(
                url=tutorial.url, directory=directory, branch=branch
            )
            html = ""
            if len(downloaded_files) == 0:
                html += f'<div class="alert alert-danger">{tutorial_name} failed to download.'
            else:
                first = downloaded_files[0]
                jupyter_path = first.replace(os.path.abspath(directory) + "/", "")

                html += f'<div class="alert alert-success">{tutorial_name} downloaded.'
                html += (
                    f'<br />📖 <a href="{jupyter_path}">Click to Open Tutorial</a></div>'
                )
            return NBOutput(html)

    def _repr_html_(self) -> str:
        html = ""
        if not arg_cache["install_wizard_complete"]:
            html += "<h3>Step 1b: Install 🧙🏽‍♂️ Wizard (Recommended)</h3>"
            html += (
                "It looks like this might be your first time running Quickstart.<br />"
            )
            html += (
                "<blockquote>Please go through the Install Wizard notebook to "
                + "install Syft and optionally start a Grid server."
            )
            html += (
                '<br />📖 <a href="./01-install-wizard.ipynb">Click to start '
                + "Install 🧙🏽‍♂️ Wizard</a></div></blockquote>"
            )
            html += "<br />"

        html += "<h3>Download Tutorials</h3>"
        html += "Below is a list of tutorials to download using quickstart.<br />"
        html += "<ul>"
        for name, tutorial in TUTORIALS.items():
            html += (
                "<li style='list-style:none;'>📖 Tutorial Series: "
                + f"<strong>{name}</strong><br />{tutorial.description}</li>"
            )
        html += "</ul>"
        first = list(TUTORIALS.keys())[0]
        html += (
            "<blockquote>Try running: <br /><code>"
            + f'quickstart.download("{first}")</code></blockquote>'
        )

        return html


def get_urls_from_dir(
    url: str,
    repo: str,
    branch: str,
    commit: Optional[str] = None,
) -> List[str]:
    notebooks = []
    slug = commit if commit else branch

    gh_api_call = (
        "https://api.github.com/repos/" + repo + "/git/trees/" + slug + "?recursive=1"
    )
    r = requests.get(gh_api_call)  # nosec
    if r.status_code != 200:
        print(
            f"Failed to fetch notebook from: {gh_api_call}.\n"
            + "Please try again with the correct parameters!"
        )
        sys.exit(1)

    res = r.json()

    for file in res["tree"]:
        if file["path"].startswith("notebooks/quickstart/" + url) or file[
            "path"
        ].startswith("notebooks/" + url):
            if file["path"].endswith(".ipynb"):
                temp_url = (
                    "https://raw.githubusercontent.com/"
                    + repo
                    + "/"
                    + slug
                    + "/"
                    + file["path"]
                )
                notebooks.append(temp_url)

    if len(notebooks) == 0:
        for file in res["tree"]:
            if file["path"].startswith("notebooks/" + url):
                if file["path"].endswith(".ipynb"):
                    temp_url = (
                        "https://raw.githubusercontent.com/"
                        + repo
                        + "/"
                        + slug
                        + "/"
                        + file["path"]
                    )
                    notebooks.append(temp_url)
    return notebooks



# stdlib
import io

# third party
import click
import rich

# relative
from .deps import DEPENDENCIES
from .mode import EDITABLE_MODE


class RichGroup(click.Group):
    def format_usage(
        self, ctx: click.core.Context, formatter: click.formatting.HelpFormatter
    ) -> None:
        sio = io.StringIO()
        console = rich.get_console()
        mode = ""
        if EDITABLE_MODE:
            mode = "[bold red]EDITABLE DEV MODE[/bold red] :police_car_light:"
        console.print(
            "[bold red]HA[/bold red][bold magenta]Grid[/bold magenta]!", ":mage:", mode
        )
        table = rich.table.Table()

        table.add_column("Dependency", style="magenta")
        table.add_column("Found", justify="right")

        for dep in sorted(DEPENDENCIES.keys()):
            path = DEPENDENCIES[dep]
            installed_str = ":white_check_mark:" if path is not None else ":cross_mark:"
            dep_emoji = ":gear:"
            if dep == "docker":
                dep_emoji = ":whale:"
            if dep == "git":
                dep_emoji = ":file_folder:"
            if dep == "ansible-playbook":
                dep_emoji = ":blue_book:"
            table.add_row(f"{dep_emoji} {dep}", installed_str)
            # console.print(dep_emoji, dep, installed_str)
        console.print(table)
        console.print("Usage: hagrid [OPTIONS] COMMAND [ARGS]...")
        formatter.write(sio.getvalue())



# stdlib
import json
import os
from typing import Any

STABLE_BRANCH = "0.8.1"
DEFAULT_BRANCH = "0.8.1"
DEFAULT_REPO = "OpenMined/PySyft"

arg_defaults = {
    "repo": DEFAULT_REPO,
    "branch": STABLE_BRANCH,
    "username": "root",
    "auth_type": "key",
    "key_path": "~/.ssh/id_rsa",
    "azure_repo": DEFAULT_REPO,
    "azure_branch": STABLE_BRANCH,
    "azure_username": "azureuser",
    "azure_key_path": "~/.ssh/id_rsa",
    "azure_resource_group": "openmined",
    "azure_location": "westus",
    "azure_size": "Standard_D4s_v3",
    "gcp_zone": "us-central1-c",
    "gcp_machine_type": "e2-standard-4",
    "gcp_project_id": "",
    "gcp_username": "",
    "gcp_key_path": "~/.ssh/google_compute_engine",
    "gcp_repo": DEFAULT_REPO,
    "gcp_branch": STABLE_BRANCH,
    "install_wizard_complete": False,
    "aws_region": "us-east-1",
    "aws_security_group_name": "openmined_sg",
    "aws_security_group_cidr": "0.0.0.0/0",
    "aws_image_id": "ami-05de688637f3e33ee",  # Ubuntu Server 22.04 LTS (HVM), SSD Volume Type
    "aws_ec2_instance_type": "t2.xlarge",
    "aws_ec2_instance_username": "ubuntu",  # For Ubuntu AMI, the default user name is ubuntu
    "aws_repo": DEFAULT_REPO,
    "aws_branch": STABLE_BRANCH,
}


class ArgCache:
    @staticmethod
    def cache_file_path() -> str:
        dir_path = os.path.expanduser("~/.hagrid")
        os.makedirs(dir_path, exist_ok=True)
        return f"{dir_path}/cache.json"

    def __init__(self) -> None:
        try:
            with open(ArgCache.cache_file_path()) as f:
                self.__cache = json.loads(f.read())
        except Exception:  # nosec
            self.__cache = {}

    def __setitem__(self, key: str, value: Any) -> None:
        self.__cache[key] = value
        with open(ArgCache.cache_file_path(), "w") as f:
            f.write(json.dumps(self.__cache))

    def __getitem__(self, key: str) -> Any:
        if key in self.__cache:
            return self.__cache[key]
        elif key in arg_defaults:
            return arg_defaults[key]
        raise KeyError(f"Can't find key {key} in ArgCache")


arg_cache = ArgCache()
